import torch
import torch.nn as nn
import math
import torch.nn.functional as ff


class OrthogonalRFFLayer(nn.Module):
    def __init__(self,
                 in_size: int,
                 out_size: int,
                 sigma: float,
                 std: float,
                 bias=True,
                 cos=True):
        super(OrthogonalRFFLayer, self).__init__()

        self.in_size = in_size
        self.out_size = out_size
        self.bias = bias
        self.std = std
        self.cos = cos

        self.register_buffer('random_layer_weight', torch.empty(out_size, in_size, requires_grad=False))

        self.register_buffer('random_layer_bias', torch.empty(out_size, requires_grad=False))

        self.scaling_factor_fix = math.sqrt(2 * sigma**2 / out_size)

        self.reset_params()

    def reset_params(self):
        with torch.no_grad():
            if self.in_size < self.out_size:
                remaining = self.out_size
                parts = []
                while remaining > 0:
                    weight = torch.randn(self.in_size, self.in_size) * self.std
                    q, r = torch.linalg.qr(weight)
                    dr = torch.diag(torch.diagonal(r))
                    q = torch.matmul(q, torch.sign(dr))
                    parts.append(q)
                    remaining = remaining - self.in_size
                weight = torch.concat(parts, dim=0)[:self.out_size, :]
            else:
                weight = torch.randn(self.in_size, self.in_size) * self.std
                q, r = torch.linalg.qr(weight)
                dr = torch.diag(torch.diagonal(r))
                weight = torch.matmul(q, torch.sign(dr))
                weight = weight[:self.out_size, :self.in_size]

            feature_norms = torch.randn_like(weight).norm(p='fro', dim=1)
            weight = torch.mul(weight,  feature_norms.unsqueeze(1))

            self.random_layer_weight = weight
            self.random_layer_weight.requires_grad = False
            if self.bias:
                self.random_layer_bias = torch.rand_like(self.random_layer_bias) * 2 * math.pi
                self.random_layer_bias.requires_grad = False

    def forward(self, x):
        if self.cos:
            return self.scaling_factor_fix * torch.cos(torch.matmul(self.random_layer_weight, x.unsqueeze(-1)).squeeze(-1) + self.random_layer_bias.unsqueeze(0))
        else:
            x = torch.matmul(self.random_layer_weight, x.unsqueeze(-1)).squeeze(-1) + self.random_layer_bias.unsqueeze(0)
            return (torch.cos(x) * (1 + (0.01 * x)**2)) * self.scaling_factor_fix


class RFFLayer(nn.Module):
    def __init__(self,
                 in_size: int,
                 out_size: int,
                 sigma: float,
                 std: float,
                 bias=True,
                 cos=True):
        super(RFFLayer, self).__init__()

        self.in_size = in_size
        self.out_size = out_size
        self.bias = bias
        self.std = std
        self.cos = cos

        self.random_layer = nn.Linear(in_features=in_size,
                                      out_features=out_size,
                                      bias=bias)

        self.scaling_factor_fix = math.sqrt(2 * sigma**2 / out_size)

        self.reset_params()

    def reset_params(self):
        # nn.init.normal_(self.random_layer.weight, mean=0.0, std=self.std)
        nn.init.trunc_normal_(self.random_layer.weight, mean=0.0, std=self.std, a=- 2.0, b=2.0)
        self.random_layer.weight.requires_grad = False
        if self.bias:
            nn.init.uniform_(self.random_layer.bias, a=0.0, b=2 * math.pi)
            self.random_layer.bias.requires_grad = False

    def forward(self, x):
        if self.cos:
            return self.scaling_factor_fix * torch.cos(self.random_layer(x))
        else:
            return self.scaling_factor_fix * self.random_layer(x)


class SNGP_Classification_simple(nn.Module):
    def __init__(self,
                 in_size: int,
                 out_size: int,
                 random_feature_size: int,
                 gp_output_bias=True,
                 normalize_input=True,
                 gp_kernel_scale=1.0,
                 gp_cov_ridge_penalty=1e-6,
                 sigma=1.0,
                 std=1.0,
                 rff_bias=True):
        super(SNGP_Classification_simple, self).__init__()
        self.in_size = in_size
        self.out_size = out_size
        self.random_feature_size = random_feature_size
        self.gp_output_bias = gp_output_bias
        self.gp_input_scale = 1. / math.sqrt(gp_kernel_scale)
        self.normalize_input = normalize_input
        self.gp_cov_ridge_penalty = gp_cov_ridge_penalty

        self.rff = OrthogonalRFFLayer(in_size=in_size,
                                      out_size=random_feature_size,
                                      sigma=sigma,
                                      std=std,
                                      bias=rff_bias)
        # self.rff = RFFLayer(in_size=in_size,
        #                     out_size=random_feature_size,
        #                     sigma=sigma,
        #                     std=std,
        #                     bias=rff_bias)
        self.beta_layer = nn.Linear(in_features=random_feature_size,
                                    out_features=out_size,
                                    bias=gp_output_bias)
        if normalize_input:
            self.laynorm = nn.LayerNorm(self.in_size)
        self.precision_matrix = None
        self.population_count = 0
        self.covariance_matrix = None

        self.FINALIZE = False

        self.reset_params()

    def reset_params(self):
        with torch.no_grad():
            # nn.init.normal_(self.beta_layer.weight, mean=0.0, std=1.0)
            nn.init.xavier_normal_(self.beta_layer.weight, gain=1.0)
            if self.gp_output_bias:
                nn.init.zeros_(self.beta_layer.bias)

    def finalize(self):
        self.train(False)
        self.FINALIZE = True
        delattr(self, 'precision_matrix')
        self.register_buffer('precision_matrix', torch.zeros(self.random_feature_size, self.random_feature_size, device=self.beta_layer.weight.device))
        self.population_count = 0
        return self

    def train(self, mode=True):
        r"""Sets the module in training mode.
        Overloaded to handle FINALIZE"""
        self.training = mode
        self.FINALIZE = False
        for module in self.children():
            module.train(mode)
        return self

    def eval(self):
        r"""Sets the module in evaluation mode.
        Overloaded to handle FINALIZE"""
        self.FINALIZE = False
        return self.train(False)

    def forward(self, x):
        valid_flag_combination = [[True, False], [False, True], [False, False]]
        flag_combination = [self.training is True, self.FINALIZE is True]
        assert flag_combination in valid_flag_combination

        if self.FINALIZE:
            return self.forward_finalize(x)
        elif self.training:
            return self.forward_training(x)
        else:
            return self.forward_eval(x)

    def forward_training(self, x):
        if self.normalize_input:
            x = self.laynorm(x)
        x = x * self.gp_input_scale
        x = self.rff(x)
        x = self.beta_layer(x)

        return x

    def forward_eval(self, x):
        if self.normalize_input:
            x = self.laynorm(x)
        x = x * self.gp_input_scale
        x_gp_features = self.rff(x)
        x = self.beta_layer(x_gp_features)
        cov = self.compute_predictive_covariance(x_gp_features)
        return x, cov

    def forward_finalize(self, x):
        if self.normalize_input:
            x = self.laynorm(x)
        x = x * self.gp_input_scale
        x_gp_features = self.rff(x)
        # in this case you need the probabilities (for now only the maximum)
        # max_probabilities, _ = torch.max(ff.softmax(self.beta_layer(x_gp_features), dim=-1), dim=-1)
        # self.precision_matrix += torch.sum((max_probabilities * (1-max_probabilities))[:, None, None] * torch.bmm(x_gp_features.unsqueeze(2), x_gp_features.unsqueeze(1)), dim=0)
        self.precision_matrix += torch.sum(torch.bmm(x_gp_features.unsqueeze(2), x_gp_features.unsqueeze(1)), dim=0)
        self.population_count += x_gp_features.shape[0]
        return None

    def gen_laplace_approx(self):
        assert self.precision_matrix is not None
        delattr(self, 'covariance_matrix')
        self.precision_matrix = self.precision_matrix / self.population_count + torch.eye(*self.precision_matrix.shape, device=self.precision_matrix.device) * self.gp_cov_ridge_penalty
        # self.precision_matrix = self.precision_matrix + torch.eye(*self.precision_matrix.shape, device=self.precision_matrix.device)
        # self.precision_matrix = self.precision_matrix + torch.eye(*self.precision_matrix.shape, device=self.precision_matrix.device) * self.gp_cov_ridge_penalty
        # self.precision_matrix = self.precision_matrix / self.population_count
        # self.precision_matrix.fill_diagonal_(1.0)
        self.register_buffer('covariance_matrix', torch.linalg.inv(self.precision_matrix))
        # determine the condition number k
        k = torch.linalg.matrix_norm(self.precision_matrix, ord=2) * torch.linalg.matrix_norm(self.covariance_matrix, ord=2)
        print('condition number k: {}'.format(k))
        self.precision_matrix = None
        self.FINALIZE = False

    def compute_predictive_covariance(self, gp_feature):
        assert self.covariance_matrix is not None
        # Computes the covariance matrix of the gp prediction.
        cov_feature_product = torch.matmul(self.covariance_matrix, gp_feature.unsqueeze(-1)) * self.gp_cov_ridge_penalty
        gp_cov_matrix = torch.matmul(gp_feature.unsqueeze(1), cov_feature_product)
        return gp_cov_matrix.view(-1)
