import torch
import numpy as np
from typing import Optional


class BatchIndexerShuffle:

    def __init__(self, ds_size: Optional[int], subset: Optional[torch.Tensor], b_size: int, device: torch.device, drop_last: bool):
        super(BatchIndexerShuffle, self).__init__()

        valid_inp = [[True, False], [False, True]]
        assert [ds_size is None, subset is None] in valid_inp

        if subset is not None:
            assert isinstance(subset, torch.Tensor)
            assert len(subset.shape) == 1
            self.subset = subset.to(device)
            self.ds_size = len(subset)
        else:
            self.subset = None
            self.ds_size = ds_size
        self.b_size = b_size

        self.device_ = device
        self.drop_last = drop_last
        self.current_permutation = None

    def generate_b_idx(self):
        if self.subset is not None:
            self.current_permutation = self.subset[torch.randperm(self.ds_size, device=self.device_)]
        else:
            self.current_permutation = torch.randperm(self.ds_size, device=self.device_)

        num_full_batches = int(np.floor(self.ds_size / self.b_size))

        batch_ = 0

        for batch in range(num_full_batches):
            batch_ = batch
            yield self.current_permutation[batch * self.b_size:(batch + 1) * self.b_size]
        # print("Full Batches Done")
        if not self.drop_last:
            yield self.current_permutation[(batch_ + 1) * self.b_size:]


class BatchIndexerSeq:

    def __init__(self, ds_size: Optional[int], subset: Optional[torch.Tensor], b_size: int, device: torch.device, drop_last: bool):
        super(BatchIndexerSeq, self).__init__()

        valid_inp = [[True, False], [False, True]]
        assert [ds_size is None, subset is None] in valid_inp

        if subset is not None:
            assert isinstance(subset, torch.Tensor)
            assert len(subset.shape) == 1
            self.sequential_idxs = subset.to(device)
        else:
            self.sequential_idxs = torch.arange(ds_size, device=device)
        self.ds_size = len(self.sequential_idxs)
        self.b_size = b_size

        self.device_ = device
        self.drop_last = drop_last
        self.current_permutation = None

    def generate_b_idx(self):
        num_full_batches = int(np.floor(self.ds_size / self.b_size))

        batch_ = 0

        for batch in range(num_full_batches):
            batch_ = batch
            yield self.sequential_idxs[batch * self.b_size:(batch + 1) * self.b_size]
        # print("Full Batches Done")
        if not self.drop_last:
            yield self.sequential_idxs[(batch_ + 1) * self.b_size:]
